@@ -2,7 +2,8 @@
 /* vim:set et sts=4: */
 /* ibus - The Input Bus
  * Copyright (C) 2008-2013 Peng Huang <shawn.p.huang@gmail.com>
- * Copyright (C) 2008-2013 Red Hat, Inc.
+ * Copyright (C) 2015-2017 Takao Fujiwara <takao.fujiwara1@gmail.com>
+ * Copyright (C) 2008-2017 Red Hat, Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -30,6 +31,10 @@
 #include <ibus.h>
 #include "ibusimcontext.h"
 
+#ifdef GDK_WINDOWING_WAYLAND
+#include <gdk/gdkwayland.h>
+#endif
+
 #if !GTK_CHECK_VERSION (2, 91, 0)
 #  define DEPRECATED_GDK_KEYSYMS 1
 #endif
@@ -86,6 +91,9 @@ static guint    _key_snooper_id = 0;
 
 static gboolean _use_sync_mode = FALSE;
 
+static const gchar *_discard_password_apps  = "";
+static gboolean _use_discard_password = FALSE;
+
 static GtkIMContext *_focus_im_context = NULL;
 static IBusInputContext *_fake_context = NULL;
 static GdkWindow *_input_window = NULL;
@@ -152,7 +160,7 @@ static gboolean _slave_delete_surrounding_cb
                                              IBusIMContext      *context);
 static void     _request_surrounding_text   (IBusIMContext      *context);
 static void     _create_fake_input_context  (void);
-static void     _set_content_type           (IBusIMContext      *context);
+static gboolean _set_content_type           (IBusIMContext      *context);
 
 
 
@@ -243,6 +251,51 @@ _focus_out_cb (GtkWidget     *widget,
     return FALSE;
 }
 
+static gboolean
+ibus_im_context_commit_event (IBusIMContext *ibusimcontext,
+                              GdkEventKey   *event)
+{
+    int i;
+    GdkModifierType no_text_input_mask;
+    gunichar ch;
+
+    if (event->type == GDK_KEY_RELEASE)
+        return FALSE;
+    /* Ignore modifier key presses */
+    for (i = 0; i < G_N_ELEMENTS (IBUS_COMPOSE_IGNORE_KEYLIST); i++)
+        if (event->keyval == IBUS_COMPOSE_IGNORE_KEYLIST[i])
+            return FALSE;
+#if GTK_CHECK_VERSION (3, 4, 0)
+    no_text_input_mask = gdk_keymap_get_modifier_mask (
+            gdk_keymap_get_for_display (gdk_display_get_default ()),
+            GDK_MODIFIER_INTENT_NO_TEXT_INPUT);
+#else
+#  ifndef GDK_WINDOWING_QUARTZ
+#    define _IBUS_NO_TEXT_INPUT_MOD_MASK (GDK_MOD1_MASK | GDK_CONTROL_MASK)
+#  else
+#    define _IBUS_NO_TEXT_INPUT_MOD_MASK (GDK_MOD2_MASK | GDK_CONTROL_MASK)
+#  endif
+
+    no_text_input_mask = _IBUS_NO_TEXT_INPUT_MOD_MASK;
+
+#  undef _IBUS_NO_TEXT_INPUT_MOD_MASK
+#endif
+    if (event->state & no_text_input_mask ||
+        event->keyval == GDK_KEY_Return ||
+        event->keyval == GDK_KEY_ISO_Enter ||
+        event->keyval == GDK_KEY_KP_Enter) {
+        return FALSE;
+    }
+    ch = ibus_keyval_to_unicode (event->keyval);
+    if (ch != 0 && !g_unichar_iscntrl (ch)) {
+        IBusText *text = ibus_text_new_from_unichar (ch);
+        g_signal_emit (ibusimcontext, _signal_commit_id, 0, text->text);
+        g_object_unref (text);
+        return TRUE;
+    }
+   return FALSE;
+}
+
 static void
 _process_key_event_done (GObject      *object,
                          GAsyncResult *res,
@@ -333,7 +386,7 @@ _request_surrounding_text (IBusIMContext *context)
     }
 }
 
-static void
+static gboolean
 _set_content_type (IBusIMContext *context)
 {
 #if GTK_CHECK_VERSION (3, 6, 0)
@@ -346,11 +399,18 @@ _set_content_type (IBusIMContext *context)
                       "input-hints", &hints,
                       NULL);
 
+        if (_use_discard_password) {
+            if (purpose == GTK_INPUT_PURPOSE_PASSWORD ||
+                purpose == GTK_INPUT_PURPOSE_PIN) {
+                return FALSE;
+            }
+        }
         ibus_input_context_set_content_type (context->ibuscontext,
                                              purpose,
                                              hints);
     }
 #endif
+    return TRUE;
 }
 
 
@@ -558,29 +618,49 @@ ibus_im_context_class_init (IBusIMContextClass *class)
     _use_key_snooper = !_get_boolean_env ("IBUS_DISABLE_SNOOPER",
                                           !(ENABLE_SNOOPER));
     _use_sync_mode = _get_boolean_env ("IBUS_ENABLE_SYNC_MODE", FALSE);
+    _use_discard_password = _get_boolean_env ("IBUS_DISCARD_PASSWORD", FALSE);
+
+#define CHECK_APP_IN_CSV_ENV_VARIABLES(retval,                          \
+                                       env_apps,                        \
+                                       fallback_apps,                   \
+                                       value_if_found)                  \
+{                                                                       \
+    const gchar * prgname = g_get_prgname ();                           \
+    gchar **p;                                                          \
+    gchar ** apps;                                                      \
+    if (g_getenv ((#env_apps))) {                                       \
+        fallback_apps = g_getenv (#env_apps);                           \
+    }                                                                   \
+    apps = g_strsplit ((fallback_apps), ",", 0);                        \
+    for (p = apps; *p != NULL; p++) {                                   \
+        if (g_regex_match_simple (*p, prgname, 0, 0)) {                 \
+            retval = (value_if_found);                                  \
+            break;                                                      \
+        }                                                               \
+    }                                                                   \
+    g_strfreev (apps);                                                  \
+}
 
     /* env IBUS_DISABLE_SNOOPER does not exist */
     if (_use_key_snooper) {
         /* disable snooper if app is in _no_snooper_apps */
-        const gchar * prgname = g_get_prgname ();
-        if (g_getenv ("IBUS_NO_SNOOPER_APPS")) {
-            _no_snooper_apps = g_getenv ("IBUS_NO_SNOOPER_APPS");
-        }
-        gchar **p;
-        gchar ** apps = g_strsplit (_no_snooper_apps, ",", 0);
-        for (p = apps; *p != NULL; p++) {
-            if (g_regex_match_simple (*p, prgname, 0, 0)) {
-                _use_key_snooper = FALSE;
-                break;
-            }
-        }
-        g_strfreev (apps);
+        CHECK_APP_IN_CSV_ENV_VARIABLES (_use_key_snooper,
+                                        IBUS_NO_SNOOPER_APPS,
+                                        _no_snooper_apps,
+                                        FALSE);
     }
+    if (!_use_discard_password) {
+        CHECK_APP_IN_CSV_ENV_VARIABLES (_use_discard_password,
+                                        IBUS_DISCARD_PASSWORD_APPS,
+                                        _discard_password_apps,
+                                        TRUE);
+    }
+
+#undef CHECK_APP_IN_CSV_ENV_VARIABLES
 
     /* init bus object */
     if (_bus == NULL) {
-        ibus_set_display (gdk_display_get_name (gdk_display_get_default ()));
-        _bus = ibus_bus_new_async ();
+        _bus = ibus_bus_new_async_client ();
 
         /* init the global fake context */
         if (ibus_bus_is_connected (_bus)) {
@@ -600,7 +680,7 @@ ibus_im_context_class_init (IBusIMContextClass *class)
     }
 
     _daemon_name_watch_id = g_bus_watch_name (G_BUS_TYPE_SESSION,
-                                              IBUS_SERVICE_IBUS,
+                                              ibus_bus_get_service_name (_bus),
                                               G_BUS_NAME_WATCHER_FLAGS_NONE,
                                               daemon_name_appeared,
                                               daemon_name_vanished,
@@ -794,8 +874,11 @@ ibus_im_context_filter_keypress (GtkIMContext *context,
     if (event->state & IBUS_HANDLED_MASK)
         return TRUE;
 
+    /* Do not call gtk_im_context_filter_keypress() because
+     * gtk_im_context_simple_filter_keypress() binds Ctrl-Shift-u
+     */
     if (event->state & IBUS_IGNORED_MASK)
-        return gtk_im_context_filter_keypress (ibusimcontext->slave, event);
+        return ibus_im_context_commit_event (ibusimcontext, event);
 
     /* XXX it is a workaround for some applications do not set client
      * window. */
@@ -874,7 +957,10 @@ ibus_im_context_focus_in (GtkIMContext *context)
 
     ibusimcontext->has_focus = TRUE;
     if (ibusimcontext->ibuscontext) {
-        _set_content_type (ibusimcontext);
+        if (!_set_content_type (ibusimcontext)) {
+            ibusimcontext->has_focus = FALSE;
+            return;
+        }
         ibus_input_context_focus_in (ibusimcontext->ibuscontext);
     }
 
@@ -906,9 +992,14 @@ ibus_im_context_focus_out (GtkIMContext *context)
         return;
     }
 
-    g_object_remove_weak_pointer ((GObject *) context,
-                                  (gpointer *) &_focus_im_context);
-    _focus_im_context = NULL;
+    /* If _use_discard_password is TRUE or GtkEntry has no visibility,
+     * _focus_im_context is NULL.
+     */
+    if (_focus_im_context) {
+        g_object_remove_weak_pointer ((GObject *) context,
+                                      (gpointer *) &_focus_im_context);
+        _focus_im_context = NULL;
+    }
 
     ibusimcontext->has_focus = FALSE;
     if (ibusimcontext->ibuscontext) {
@@ -996,6 +1087,24 @@ ibus_im_context_set_client_window (GtkIMContext *context, GdkWindow *client)
         gtk_im_context_set_client_window (ibusimcontext->slave, client);
 }
 
+static void
+_set_rect_scale_factor_with_window (GdkRectangle *area,
+                                    GdkWindow    *window)
+{
+#if GTK_CHECK_VERSION (3, 10, 0)
+    int scale_factor;
+
+    g_assert (area);
+    g_assert (GDK_IS_WINDOW (window));
+
+    scale_factor = gdk_window_get_scale_factor (window);
+    area->x *= scale_factor;
+    area->y *= scale_factor;
+    area->width *= scale_factor;
+    area->height *= scale_factor;
+#endif
+}
+
 static gboolean
 _set_cursor_location_internal (IBusIMContext *ibusimcontext)
 {
@@ -1007,6 +1116,32 @@ _set_cursor_location_internal (IBusIMContext *ibusimcontext)
     }
 
     area = ibusimcontext->cursor_area;
+
+#ifdef GDK_WINDOWING_WAYLAND
+    if (GDK_IS_WAYLAND_DISPLAY (gdk_display_get_default ())) {
+        gdouble px, py;
+        GdkWindow *parent;
+        GdkWindow *window = ibusimcontext->client_window;
+
+        while ((parent = gdk_window_get_effective_parent (window)) != NULL) {
+            gdk_window_coords_to_parent (window, area.x, area.y, &px, &py);
+            area.x = px;
+            area.y = py;
+            window = parent;
+        }
+
+        _set_rect_scale_factor_with_window (&area,
+                                            ibusimcontext->client_window);
+        ibus_input_context_set_cursor_location_relative (
+            ibusimcontext->ibuscontext,
+            area.x,
+            area.y,
+            area.width,
+            area.height);
+        return FALSE;
+    }
+#endif
+
     if (area.x == -1 && area.y == -1 && area.width == 0 && area.height == 0) {
 #if GTK_CHECK_VERSION (2, 91, 0)
         area.x = 0;
@@ -1022,6 +1157,7 @@ _set_cursor_location_internal (IBusIMContext *ibusimcontext)
     gdk_window_get_root_coords (ibusimcontext->client_window,
                                 area.x, area.y,
                                 &area.x, &area.y);
+    _set_rect_scale_factor_with_window (&area, ibusimcontext->client_window);
     ibus_input_context_set_cursor_location (ibusimcontext->ibuscontext,
                                             area.x,
                                             area.y,
